/*
    Bantumi
    Copyright 2011 Martin Storsjo

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

    Martin Storsjo
    martin@martin.st
*/

#include "st_martin_bantumi_BantumiActivity.h"
#include "st_martin_bantumi_BantumiRenderer.h"
#include "jniutils.h"
#include "bantumigl.h"
#include "glfont.h"
#include <string.h>
#include <list>
#include "tcpip.h"
#include "mutex.h"

class BantumiContext;
class AndroidConnection;
class JavaConnection;

static JNIEnv* curUpdateEnv;
static BantumiContext* curUpdateCtx;
static jobject curUpdateActivity;

class BantumiContext {
public:
	BantumiContext() {
		game = NULL;
		frontEnd = NULL;
		level = num = -1;
		conn = NULL;
		javaConn = NULL;
	}
	void setLevelNum(int l, int n) {
		MutexLock lock(mutex);
		level = l;
		num = n;
		if (frontEnd) {
			frontEnd->setLevel(level);
			frontEnd->setNum(num);
		}
	}
	int getLevel() const {
		MutexLock lock(mutex);
		if (frontEnd)
			return frontEnd->getLevel();
		return level;
	}
	int getNum() const {
		MutexLock lock(mutex);
		if (frontEnd)
			return frontEnd->getNum();
		return num;
	}
	void initGL(int w, int h) {
		MutexLock lock(mutex);
		if (!frontEnd) {
			frontEnd = new BantumiGL(w, h);
			game = new Bantumi(frontEnd);
			if (level >= 0)
				frontEnd->setLevel(level);
			if (num >= 0)
				frontEnd->setNum(num);
		} else
			frontEnd->reinitGL(w, h);
	}
	~BantumiContext() {
		delete game;
		delete frontEnd;
	}
	void draw(int time, JNIEnv* env = NULL, jobject activity = NULL) {
		MutexLock lock(mutex);
		curUpdateEnv = env;
		curUpdateCtx = this;
		curUpdateActivity = activity;
		while (!clickQueue.empty()) {
			std::pair<int, int> coord = clickQueue.front();
			clickQueue.pop_front();
			game->click(coord.first, coord.second, 25);
		}
		while (!pressQueue.empty()) {
			Key k = pressQueue.front();
			pressQueue.pop_front();
			game->pressed(k);
		}
		game->update();
		frontEnd->draw(time);
		curUpdateEnv = NULL;
		curUpdateCtx = NULL;
		curUpdateActivity = NULL;
	}
	void queueClick(int x, int y) {
		MutexLock lock(mutex);
		clickQueue.push_back(std::pair<int, int>(x, y));
	}
	void queuePress(Key k) {
		MutexLock lock(mutex);
		pressQueue.push_back(k);
	}
	Bantumi* game;
	BantumiGL* frontEnd;
	std::list<std::pair<int, int> > clickQueue;
	std::list<Key> pressQueue;
	int level, num;
	AndroidConnection* conn;
	mutable Mutex mutex;
	JavaConnection* javaConn;
};

#define getBantumiPrivData getPrivData<BantumiContext*>
#define setBantumiPrivData setPrivData<BantumiContext*>

JNIEXPORT void JNICALL Java_st_martin_bantumi_BantumiRenderer_nativeOnDrawFrame(JNIEnv* env, jobject thiz, jobject activity, jlong time) {
	BantumiContext *ctx = getBantumiPrivData(env, activity);
	if (ctx && ctx->frontEnd) {
		ctx->draw(time, env, activity);
	} else {
		glClearColor(0, 0, 0, 1);
		glClear(GL_COLOR_BUFFER_BIT);
	}
}

JNIEXPORT void JNICALL Java_st_martin_bantumi_BantumiRenderer_nativeOnSurfaceCreated(JNIEnv* env, jobject thiz, jobject activity) {
}

JNIEXPORT void JNICALL Java_st_martin_bantumi_BantumiRenderer_nativeOnSurfaceChanged(JNIEnv* env, jobject thiz, jobject activity, jint width, jint height) {
	BantumiContext *ctx = getBantumiPrivData(env, activity);
	if (ctx)
		ctx->initGL(width, height);
	setBantumiPrivData(env, activity, ctx);
}

JNIEXPORT void JNICALL Java_st_martin_bantumi_BantumiRenderer_setFont(JNIEnv* env, jobject thiz, jobject activity, jboolean text, jbyteArray image, jboolean alphaBitmap, jint w, int h, jintArray info, jint offset, jint n, jint chars) {
	BantumiContext *ctx = getBantumiPrivData(env, activity);
	if (!ctx || !ctx->frontEnd)
		return;
	TexGLFont* f = new TexGLFont(offset, n);

	int* infoPtr = env->GetIntArrayElements(info, NULL);
	for (int i = 0; i < chars; i++) {
		int* charInfo = &infoPtr[8*i];
		f->setParams(charInfo[0], charInfo[1], charInfo[2], charInfo[3], charInfo[4], charInfo[5], charInfo[6], charInfo[7]);
	}
	env->ReleaseIntArrayElements(info, infoPtr, JNI_ABORT);

	jbyte* ptr = env->GetByteArrayElements(image, NULL);
	if (alphaBitmap)
		f->setAlphaImage((uint8_t*) ptr, w, h);
	else
		f->setImage((uint8_t*) ptr, w, h, 0);
	env->ReleaseByteArrayElements(image, ptr, JNI_ABORT);

	if (text)
		ctx->frontEnd->setTextFont(f);
	else
		ctx->frontEnd->setNumFont(f);

}

JNIEXPORT void JNICALL Java_st_martin_bantumi_BantumiActivity_init(JNIEnv* env, jobject thiz) {
	BantumiContext *ctx = getBantumiPrivData(env, thiz);
	delete ctx;
	ctx = new BantumiContext();
	setBantumiPrivData(env, thiz, ctx);
}

JNIEXPORT void JNICALL Java_st_martin_bantumi_BantumiActivity_close(JNIEnv* env, jobject thiz) {
	BantumiContext *ctx = getBantumiPrivData(env, thiz);
	delete ctx;
	setBantumiPrivData(env, thiz, NULL);
}

JNIEXPORT void JNICALL Java_st_martin_bantumi_BantumiActivity_handleClick(JNIEnv* env, jobject thiz, jint x, jint y) {
	BantumiContext *ctx = getBantumiPrivData(env, thiz);
	if (ctx)
		ctx->queueClick(x, y);
}

JNIEXPORT void JNICALL Java_st_martin_bantumi_BantumiActivity_handleButton(JNIEnv* env, jobject thiz, jint button) {
	BantumiContext *ctx = getBantumiPrivData(env, thiz);
	if (ctx && ctx->game)
		ctx->queuePress((Key)button);
}

JNIEXPORT jboolean JNICALL Java_st_martin_bantumi_BantumiActivity_shouldExit(JNIEnv* env, jobject thiz) {
	BantumiContext *ctx = getBantumiPrivData(env, thiz);
	if (ctx && ctx->game)
		return ctx->game->exit();
	return false;
}

JNIEXPORT void JNICALL Java_st_martin_bantumi_BantumiActivity_setLevelNum(JNIEnv* env, jobject thiz, jint level, jint num) {
	BantumiContext *ctx = getBantumiPrivData(env, thiz);
	if (ctx)
		ctx->setLevelNum(level, num);
}

JNIEXPORT jint JNICALL Java_st_martin_bantumi_BantumiActivity_getLevel(JNIEnv* env, jobject thiz) {
	BantumiContext *ctx = getBantumiPrivData(env, thiz);
	if (ctx)
		return ctx->getLevel();
	return -1;
}

JNIEXPORT jint JNICALL Java_st_martin_bantumi_BantumiActivity_getNum(JNIEnv* env, jobject thiz) {
	BantumiContext *ctx = getBantumiPrivData(env, thiz);
	if (ctx)
		return ctx->getNum();
	return -1;
}

void finishProgress() {
	JNIEnv* env = curUpdateEnv;
	jobject activity = curUpdateActivity;
	if (!env)
		return;
	jclass clazz = env->GetObjectClass(activity);
	jmethodID method = env->GetMethodID(clazz, "finishProgress", "()V");
	env->CallVoidMethod(activity, method);
	env->DeleteLocalRef(clazz);
}

class AndroidConnection : public Connection {
public:
	AndroidConnection() {
		child = NULL;
		done = false;
	}
	~AndroidConnection() {
		delete child;
		curUpdateCtx->conn = NULL;
		curUpdateCtx->javaConn = NULL;
	}
	bool isClient() {
		if (child)
			return child->isClient();
		return true;
	}
	int ready(const char **errString) {
		if (child) {
			int ret = child->ready(errString);
			if (!done && ret != 0) {
				finishProgress();
				done = true;
			}
			return ret;
		}
		return 0;
	}
	bool read(const char **errString) {
		if (child)
			return child->read(errString);
		return true;
	}
	bool write(const unsigned char *ptr, int n, const char **errString) {
		if (child)
			return child->write(ptr, n, errString);
		return false;
	}
	void setChild(Connection* conn) {
		child = conn;
		child->setCallback(cb);
	}

public:
	Connection* child;
	bool done;
};

void showWarning(JNIEnv* env, jobject activity, const char* str) {
	jstring jstr = env->NewStringUTF(str);
	jclass clazz = env->GetObjectClass(activity);
	jmethodID method = env->GetMethodID(clazz, "showWarning", "(Ljava/lang/String;)V");
	env->CallVoidMethod(activity, method, jstr);
	env->DeleteLocalRef(clazz);
	env->DeleteLocalRef(jstr);
}

void showWarning(const char* str) {
	showWarning(curUpdateEnv, curUpdateActivity, str);
}

Connection *showMultiplayerDialog() {
	JNIEnv* env = curUpdateEnv;
	jobject activity = curUpdateActivity;
	BantumiContext* ctx = curUpdateCtx;
	if (!env)
		return NULL;
	ctx->conn = new AndroidConnection();
	jclass clazz = env->GetObjectClass(activity);
	jmethodID method = env->GetMethodID(clazz, "showMultiplayerDialog", "()V");
	env->CallVoidMethod(activity, method);
	env->DeleteLocalRef(clazz);
	return ctx->conn;
}

JNIEXPORT void JNICALL Java_st_martin_bantumi_BantumiActivity_doConnect(JNIEnv* env, jobject thiz, jstring host, jint port) {
	BantumiContext *ctx = getBantumiPrivData(env, thiz);
	if (!ctx)
		return;
	MutexLock lock(ctx->mutex);
	if (!ctx->conn)
		return;
	Connection* conn = NULL;
	if (host) {
		const char* hostStr = env->GetStringUTFChars(host, NULL);
		const char* err = NULL;
		conn = TCPConnection::startConnect(hostStr, port, &err);
		if (!conn)
			showWarning(env, thiz, err);
		env->ReleaseStringUTFChars(host, hostStr);
	} else {
		const char* err = NULL;
		conn = TCPConnection::startAccept(port, &err);
		if (!conn)
			showWarning(env, thiz, err);
	}
	ctx->conn->setChild(conn);
}

void closeConnection() {
	JNIEnv* env = curUpdateEnv;
	jobject activity = curUpdateActivity;
	BantumiContext* ctx = curUpdateCtx;
	if (!env)
		return;
	jclass clazz = env->GetObjectClass(activity);
	jmethodID method = env->GetMethodID(clazz, "closeConnection", "()V");
	env->CallVoidMethod(activity, method);
	env->DeleteLocalRef(clazz);
}

void writeOutputData(const uint8_t* ptr, int n) {
	JNIEnv* env = curUpdateEnv;
	jobject activity = curUpdateActivity;
	BantumiContext* ctx = curUpdateCtx;
	if (!env)
		return;
	jbyteArray array = env->NewByteArray(n);
	jbyte* arrayPtr = env->GetByteArrayElements(array, NULL);
	memcpy(arrayPtr, ptr, n);
	env->ReleaseByteArrayElements(array, arrayPtr, 0);

	jclass clazz = env->GetObjectClass(activity);
	jmethodID method = env->GetMethodID(clazz, "writeOutputData", "([B)V");
	env->CallVoidMethod(activity, method, array);
	env->DeleteLocalRef(clazz);
	env->DeleteLocalRef(array);
}

class JavaConnection : public Connection {
public:
	JavaConnection(bool c) {
		client = c;
		status = 0;
		readBuffer = NULL;
		readBufferSize = 0;
		errMsg = NULL;
	}
	~JavaConnection() {
		closeConnection();
		free(readBuffer);
		free(errMsg);
	}
	bool isClient() {
		return client;
	}
	int ready(const char **errString) {
		*errString = NULL;
		if (status < 0)
			*errString = errMsg;
		return status;
	}
	bool read(const char **errString) {
		*errString = NULL;
		if (status < 0) {
			*errString = errMsg;
			return false;
		}
		if (readBuffer) {
			cb->received(readBuffer, readBufferSize);
			free(readBuffer);
			readBuffer = NULL;
			readBufferSize = 0;
		}
		return true;
	}
	bool write(const unsigned char *ptr, int n, const char **errString) {
		*errString = NULL;
		if (status < 0) {
			*errString = errMsg;
			return false;
		}
		writeOutputData(ptr, n);
		return true;
	}
	void setStatus(int s, const char* msg) {
		status = s;
		if (msg && !errMsg)
			errMsg = strdup(msg);
	}
	void addData(const uint8_t* buf, int n) {
		readBuffer = (uint8_t*) realloc(readBuffer, readBufferSize + n);
		memcpy(readBuffer + readBufferSize, buf, n);
		readBufferSize += n;
	}

private:
	bool client;
	int status;
	uint8_t* readBuffer;
	int readBufferSize;
	char* errMsg;

};

JNIEXPORT void JNICALL Java_st_martin_bantumi_BantumiActivity_initJavaConnection(JNIEnv* env, jobject thiz, jboolean client) {
	BantumiContext *ctx = getBantumiPrivData(env, thiz);
	if (!ctx)
		return;
	MutexLock lock(ctx->mutex);
	if (!ctx->conn)
		return;
	delete ctx->javaConn;
	ctx->javaConn = new JavaConnection(client);
	ctx->conn->setChild(ctx->javaConn);
}

JNIEXPORT void JNICALL Java_st_martin_bantumi_BantumiActivity_setConnectionStatus(JNIEnv* env, jobject thiz, jboolean working, jstring msg) {
	BantumiContext *ctx = getBantumiPrivData(env, thiz);
	if (!ctx)
		return;
	MutexLock lock(ctx->mutex);
	if (!ctx->javaConn)
		return;
	const char* msgPtr = NULL;
	if (msg)
		msgPtr = env->GetStringUTFChars(msg, NULL);
	ctx->javaConn->setStatus(working ? 1 : -1, msgPtr);
	if (msg)
		env->ReleaseStringUTFChars(msg, msgPtr);
}

JNIEXPORT void JNICALL Java_st_martin_bantumi_BantumiActivity_handleInputData(JNIEnv* env, jobject thiz, jbyteArray buffer, jint n) {
	BantumiContext *ctx = getBantumiPrivData(env, thiz);
	if (!ctx)
		return;
	MutexLock lock(ctx->mutex);
	if (!ctx->javaConn)
		return;
	jbyte* ptr = env->GetByteArrayElements(buffer, NULL);
	ctx->javaConn->addData((const uint8_t*) ptr, n);
	env->ReleaseByteArrayElements(buffer, ptr, JNI_ABORT);
}

