/*
    Bantumi
    Copyright 2005 - 2007 Martin Storsjo

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

    Martin Storsjo
    martin@martin.st
*/

#include "sdprecord.h"
#include <string.h>
#include <stdio.h>

#define SDP_TYPE_NIL	0
#define SDP_TYPE_UINT	1
#define SDP_TYPE_SINT	2
#define SDP_TYPE_UUID	3
#define SDP_TYPE_STRING	4
#define SDP_TYPE_BOOL	5
#define SDP_TYPE_DES	6
#define SDP_TYPE_DEA	7
#define SDP_TYPE_URL	8

#define SDP_TYPE_SHIFT	3

#define SDP_SIZE_1	0
#define SDP_SIZE_2	1
#define SDP_SIZE_4	2
#define SDP_SIZE_8	3
#define SDP_SIZE_16	4
#define SDP_SIZE_8BIT	5
#define SDP_SIZE_16BIT	6
#define SDP_SIZE_32BIT	7

#ifndef SDP_ATTRIB_CLASS_ID_LIST
#define SDP_ATTRIB_CLASS_ID_LIST 0x0001
#endif
#ifndef SDP_ATTRIB_PROTOCOL_DESCRIPTOR_LIST
#define SDP_ATTRIB_PROTOCOL_DESCRIPTOR_LIST 0x0004
#endif
#ifndef L2CAP_PROTOCOL_UUID16
#define L2CAP_PROTOCOL_UUID16 0x0100
#endif
#ifndef RFCOMM_PROTOCOL_UUID16
#define RFCOMM_PROTOCOL_UUID16 0x0003
#endif


static const unsigned char baseUuid[] = {
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x10, 0x00, 0x80, 0x00, 0x00, 0x80, 0x5F, 0x9B, 0x34, 0xFB
};

SdpRecord::SdpRecord() {
	contexts.push(Context(200));
}

SdpRecord::~SdpRecord() {
}

SdpRecord* SdpRecord::addNil() {
	writeType(SDP_TYPE_NIL, 0);
	return this;
}

SdpRecord* SdpRecord::addUint8(unsigned char val) {
	writeType(SDP_TYPE_UINT, SDP_SIZE_1);
	writeUint8(val);
	return this;
}

SdpRecord* SdpRecord::addUint16(unsigned short val) {
	writeType(SDP_TYPE_UINT, SDP_SIZE_2);
	writeUint16(val);
	return this;
}

SdpRecord* SdpRecord::addUint32(unsigned int val) {
	writeType(SDP_TYPE_UINT, SDP_SIZE_4);
	writeUint32(val);
	return this;
}

SdpRecord* SdpRecord::addUint64(unsigned long long val) {
	writeType(SDP_TYPE_UINT, SDP_SIZE_8);
	writeUint32((unsigned int) ((val >> 32) & 0xffffffff));
	writeUint32((unsigned int) ((val >>  0) & 0xffffffff));
	return this;
}

SdpRecord* SdpRecord::addUint(unsigned long long val) {
	if (val <= 0xFF)
		return addUint8((unsigned char) val);
	else if (val <= 0xFFFF)
		return addUint16((unsigned short) val);
	else if (val <= 0xFFFFFFFF)
		return addUint32((unsigned int) val);
	else
		return addUint64(val);
}

SdpRecord* SdpRecord::addSint8(signed char val) {
	writeType(SDP_TYPE_SINT, SDP_SIZE_1);
	writeUint8(val);
	return this;
}

SdpRecord* SdpRecord::addSint16(signed short val) {
	writeType(SDP_TYPE_SINT, SDP_SIZE_2);
	writeUint16(val);
	return this;
}

SdpRecord* SdpRecord::addSint32(signed int val) {
	writeType(SDP_TYPE_SINT, SDP_SIZE_4);
	writeUint32(val);
	return this;
}

SdpRecord* SdpRecord::addSint64(signed long long val) {
	writeType(SDP_TYPE_SINT, SDP_SIZE_8);
	writeUint32((unsigned int) ((val >> 32) & 0xffffffff));
	writeUint32((unsigned int) ((val >>  0) & 0xffffffff));
	return this;
}

SdpRecord* SdpRecord::addSint(signed long long val) {
	if (-0x80 <= val && val <= 0x7F)
		return addSint8((signed char) val);
	else if (-0x8000 <= val && val <= 0x7FFF)
		return addSint16((signed short) val);
	else if (-0x80000000LL <= val && val <= 0x7FFFFFFF)
		return addSint32((signed int) val);
	else
		return addSint64(val);
}

SdpRecord* SdpRecord::addUuid16(unsigned short uuid) {
	writeType(SDP_TYPE_UUID, SDP_SIZE_2);
	writeUint16(uuid);
	return this;
}

SdpRecord* SdpRecord::addUuid32(unsigned int uuid) {
	writeType(SDP_TYPE_UUID, SDP_SIZE_4);
	writeUint32(uuid);
	return this;
}

SdpRecord* SdpRecord::addUuid128(const unsigned char* uuid) {
	writeType(SDP_TYPE_UUID, SDP_SIZE_16);
	top().writeBytes(uuid, 16);
	return this;
}

SdpRecord* SdpRecord::addString(const char* str) {
	writeData(SDP_TYPE_STRING, (const unsigned char*) str, strlen(str));
	return this;
}

SdpRecord* SdpRecord::addBoolean(bool val) {
	writeType(SDP_TYPE_BOOL, SDP_SIZE_1);
	writeUint8(val ? 1 : 0);
	return this;
}

SdpRecord* SdpRecord::addDES() {
	contexts.push(Context(200));
	return this;
}

SdpRecord* SdpRecord::endDES() {
	Context child = top();
	contexts.pop();
	writeData(SDP_TYPE_DES, child.getData(), child.getLength());
	return this;
}

SdpRecord* SdpRecord::addDEA() {
	contexts.push(Context(200));
	return this;
}

SdpRecord* SdpRecord::endDEA() {
	Context child = top();
	contexts.pop();
	writeData(SDP_TYPE_DEA, child.getData(), child.getLength());
	return this;
}

SdpRecord* SdpRecord::addUrl(const char* url) {
	writeData(SDP_TYPE_URL, (const unsigned char*) url, strlen(url));
	return this;
}

SdpRecord* SdpRecord::startRecord() {
	return addDES();
}

SdpRecord* SdpRecord::endRecord() {
	return endDES();
}

SdpRecord* SdpRecord::addAttribute(unsigned int id) {
	return addUint16(id);
}

SdpRecord* SdpRecord::endAttribute() {
	return this;
}

const unsigned char* SdpRecord::getData() {
	return top().getData();
}

int SdpRecord::getLength() {
	return top().getLength();
}

SdpRecord::Context& SdpRecord::top() {
	return contexts.top();
}

void SdpRecord::writeType(int type, int size) {
	writeUint8((type << SDP_TYPE_SHIFT) | size);
}

void SdpRecord::writeUint8(unsigned char value) {
	top().writeByte(value);
}

void SdpRecord::writeUint16(unsigned short value) {
	top().writeByte((value >> 8) & 0xff);
	top().writeByte((value >> 0) & 0xff);
}

void SdpRecord::writeUint32(unsigned int value) {
	top().writeByte((value >> 24) & 0xff);
	top().writeByte((value >> 16) & 0xff);
	top().writeByte((value >> 8) & 0xff);
	top().writeByte((value >> 0) & 0xff);
}

void SdpRecord::writeData(int type, const unsigned char* data, int len) {
	if (len <= 0xff) {
		writeType(type, SDP_SIZE_8BIT);
		writeUint8(len);
	} else if (len <= 0xffff) {
		writeType(type, SDP_SIZE_16BIT);
		writeUint16(len);
	} else {
		writeType(type, SDP_SIZE_32BIT);
		writeUint32(len);
	}
	top().writeBytes(data, len);
}



SdpRecord::Context::Context(int l) {
	array = new unsigned char[l];
	pos = 0;
	len = l;
}

SdpRecord::Context::Context(const Context& c) {
	pos = c.pos;
	len = c.len;
	array = new unsigned char[len];
	memcpy(array, c.array, len);
}

SdpRecord::Context& SdpRecord::Context::operator=(const Context& c) {
	pos = c.pos;
	len = c.len;
	delete [] array;
	array = new unsigned char[len];
	memcpy(array, c.array, len);
	return *this;
}

SdpRecord::Context::~Context() {
	delete [] array;
}

void SdpRecord::Context::assureSpace(int s) {
	if (pos + s <= len)
		return;
	while (pos + s > len) {
		len *= 2;
	}
	unsigned char* newarray = new unsigned char[len];
	memcpy(newarray, array, pos);
	delete [] array;
	array = newarray;
}

void SdpRecord::Context::writeByte(unsigned char b) {
	assureSpace(1);
	array[pos++] = b;
}

void SdpRecord::Context::writeBytes(const unsigned char* bytes, int len) {
	assureSpace(len);
	memcpy(&array[pos], bytes, len);
	pos += len;
}

const unsigned char* SdpRecord::Context::getData() {
	return array;
}

int SdpRecord::Context::getLength() {
	return pos;
}




SdpRecordParser::Buffer::Buffer(const unsigned char* data, int len) {
	this->data = data;
	this->length = len;
}

int SdpRecordParser::Buffer::getLength() const {
	return length;
}

const unsigned char* SdpRecordParser::Buffer::getData() const {
	return data;
}

void SdpRecordParser::Buffer::consume(int bytes) {
	if (bytes > length)
		bytes = length;
	data += bytes;
	length -= bytes;
}

unsigned char SdpRecordParser::Buffer::readByte() {
	unsigned char val = *data;
	consume(1);
	return val;
}

SdpRecordParser::Buffer SdpRecordParser::Buffer::readSubstring(int bytes) {
	if (bytes > length)
		bytes = length;
	const unsigned char* ptr = data;
	consume(bytes);
	return Buffer(ptr, bytes);
}


SdpRecordParser::ArrayRecord::ArrayRecord(int type) : Record(type) {
}

SdpRecordParser::ArrayRecord::~ArrayRecord() {
	for (unsigned int i = 0; i < children.size(); i++)
		delete children[i];
}

void SdpRecordParser::ArrayRecord::appendRecord(Record* record) {
	children.push_back(record);
}

int SdpRecordParser::ArrayRecord::getCount() const {
	return children.size();
}

SdpRecordParser::Record* SdpRecordParser::ArrayRecord::getRecord(int index) const {
	if (index < 0 || index >= getCount())
		return NULL;
	return children[index];
}

SdpRecordParser::Record::Record(int type, int size) {
	this->type = type;
	this->size = size;
}

int SdpRecordParser::Record::getType() const {
	return type;
}

int SdpRecordParser::Record::getSize() const {
	return size;
}

SdpRecordParser::NilRecord::NilRecord() : Record(SDP_TYPE_NIL) {
}

SdpRecordParser::StringRecord::StringRecord() : Record(SDP_TYPE_STRING) {
	string = new char[1];
	string[0] = '\0';
}

SdpRecordParser::StringRecord::StringRecord(int type, Buffer& buffer) : Record(type) {
	int length = buffer.getLength();
	string = new char[length+1];
	memcpy(string, buffer.getData(), length);
	string[length] = '\0';
	buffer.consume(length);
}
SdpRecordParser::StringRecord::~StringRecord() {
	delete [] string;
}
const char* SdpRecordParser::StringRecord::getString() const {
	return string;
}

void SdpRecordParser::StringRecord::set(int type, const char* str) {
	this->type = type;
	delete [] string;
	int length = strlen(str);
	string = new char[length+1];
	memcpy(string, str, length);
	string[length] = '\0';
}


SdpRecordParser::PrimitiveRecord::PrimitiveRecord() : Record(SDP_TYPE_UINT, SDP_SIZE_1) {
	value = 0;
}

SdpRecordParser::PrimitiveRecord::PrimitiveRecord(int type, int size, unsigned long long value) : Record(type, size) {
	this->value = value;
}

unsigned long long SdpRecordParser::PrimitiveRecord::getValue() const {
	return value;
}

SdpRecordParser::UuidRecord::UuidRecord() : Record(SDP_TYPE_UUID, SDP_SIZE_16) {
	uuid = 0;
	memset(uuid128, 0, 16);
}

SdpRecordParser::UuidRecord::UuidRecord(int size, unsigned int uuid) : Record(SDP_TYPE_UUID, size) {
	this->uuid = uuid;
	memcpy(uuid128, baseUuid, 16);
	uuid128[0] = uuid >> 24;
	uuid128[1] = uuid >> 16;
	uuid128[2] = uuid >> 8;
	uuid128[3] = uuid >> 0;
}

SdpRecordParser::UuidRecord::UuidRecord(int size, Buffer& buffer) : Record(SDP_TYPE_UUID, size) {
	if (buffer.getLength() < 16)
		return;
	memcpy(uuid128, buffer.getData(), 16);
	if (!memcmp(&uuid128[4], &baseUuid[4], 12)) {
		uuid = (uuid128[0]<<24) | (uuid128[1]<<16) | (uuid128[2]<<8) | (uuid128[3]);
	} else {
		uuid = ~0;
	}
}

bool SdpRecordParser::UuidRecord::matches(const UuidRecord& uuid) const {
	if (!memcmp(uuid128, uuid.uuid128, 16))
		return true;
	return false;
}




SdpRecordParser::SdpRecordParser(const unsigned char* data, int length) {
	error = false;
	Buffer buffer(data, length);
	rootRecord = parseRecord(buffer);
	if (buffer.getLength() > 0)
		raiseError();
	if (error) {
		printf("SdpRecordParser found errors\n");
	}
}

SdpRecordParser::~SdpRecordParser() {
	delete rootRecord;
}

void SdpRecordParser::raiseError() {
	error = true;
}

unsigned char SdpRecordParser::readUint8(Buffer& buffer) {
	return buffer.readByte();
}

unsigned short SdpRecordParser::readUint16(Buffer& buffer) {
	if (buffer.getLength() < 2) {
		raiseError();
		return 0;
	}
	const unsigned char* data = buffer.getData();
	buffer.consume(2);
	return (data[0]<<8) | (data[1]);
}

unsigned int SdpRecordParser::readUint32(Buffer& buffer) {
	if (buffer.getLength() < 4) {
		raiseError();
		return 0;
	}
	const unsigned char* data = buffer.getData();
	buffer.consume(4);
	return (data[0]<<24) | (data[1]<<16) | (data[2]<<8) | (data[3]);
}

unsigned long long SdpRecordParser::readUint64(Buffer& buffer) {
	unsigned int high = readUint32(buffer);
	unsigned int low = readUint32(buffer);
	return (((unsigned long long)high)<<32) | low;
}

SdpRecordParser::Record* SdpRecordParser::parseRecord(Buffer& buffer) {
	if (buffer.getLength() < 1)
		return NULL;
	unsigned char header = buffer.readByte();
	int type = header >> SDP_TYPE_SHIFT;
	int size = header & ~((~0) << SDP_TYPE_SHIFT);

	if (type == SDP_TYPE_NIL)
		return new NilRecord();

	int bytes = 0;
	switch (size) {
	case SDP_SIZE_1:
		bytes = 1;
		break;
	case SDP_SIZE_2:
		bytes = 2;
		break;
	case SDP_SIZE_4:
		bytes = 4;
		break;
	case SDP_SIZE_8:
		bytes = 8;
		break;
	case SDP_SIZE_16:
		bytes = 16;
		break;
	case SDP_SIZE_8BIT:
		bytes = 1;
		break;
	case SDP_SIZE_16BIT:
		bytes = 2;
		break;
	case SDP_SIZE_32BIT:
		bytes = 4;
		break;
	}

	if (buffer.getLength() < bytes) {
		raiseError();
		return NULL;
	}
	Buffer contentBuffer = buffer.readSubstring(bytes);

	bool newContent = false;
	switch (size) {
	case SDP_SIZE_8BIT:
		newContent = true;
		bytes = readUint8(contentBuffer);
		break;
	case SDP_SIZE_16BIT:
		newContent = true;
		bytes = readUint16(contentBuffer);
		break;
	case SDP_SIZE_32BIT:
		newContent = true;
		bytes = readUint32(contentBuffer);
		break;
	}
	if (newContent) {
		if (buffer.getLength() < bytes) {
			raiseError();
			return NULL;
		}
		contentBuffer = buffer.readSubstring(bytes);
	}

	switch (type) {
	case SDP_TYPE_DES:
	case SDP_TYPE_DEA: {
		ArrayRecord* array = new ArrayRecord(type);
		while (contentBuffer.getLength() > 0 && !error) {
			Record* r = parseRecord(contentBuffer);
			array->appendRecord(r);
		}
		return array;
	}
	case SDP_TYPE_STRING:
	case SDP_TYPE_URL:
		return new StringRecord(type, contentBuffer);
	case SDP_TYPE_UINT:
	case SDP_TYPE_SINT:
	case SDP_TYPE_BOOL:
		switch (size) {
		case SDP_SIZE_1:
			return new PrimitiveRecord(type, size, contentBuffer.readByte());
		case SDP_SIZE_2:
			return new PrimitiveRecord(type, size, readUint16(contentBuffer));
		case SDP_SIZE_4:
			return new PrimitiveRecord(type, size, readUint32(contentBuffer));
		case SDP_SIZE_8:
			return new PrimitiveRecord(type, size, readUint64(contentBuffer));
		}
	case SDP_TYPE_UUID:
		switch (size) {
		case SDP_SIZE_2:
			return new UuidRecord(SDP_SIZE_2, readUint16(contentBuffer));
		case SDP_SIZE_4:
			return new UuidRecord(SDP_SIZE_4, readUint32(contentBuffer));
		case SDP_SIZE_16:
			return new UuidRecord(SDP_SIZE_16, contentBuffer);
		}
	}

	raiseError();
	return NULL;
}

SdpRecordParser::Record* SdpRecordParser::findAttribute(unsigned int attribute) const {
	if (!rootRecord)
		return NULL;
	if (rootRecord->getType() != SDP_TYPE_DES)
		return NULL;
	ArrayRecord* array = (ArrayRecord*) rootRecord;
	if (array->getCount() < 1)
		return NULL;
	while (array->getRecord(0)->getType() == SDP_TYPE_DES) {
		array = (ArrayRecord*) array->getRecord(0);
		if (array->getCount() < 1)
			return NULL;
	}
	for (int i = 0; i + 1 < array->getCount(); i += 2) {
		Record* key = array->getRecord(i);
		Record* value = array->getRecord(i + 1);
		switch (key->getType()) {
		case SDP_TYPE_UINT:
		case SDP_TYPE_SINT:
			break;
		default:
			// bad key type
			return NULL;
		}
		PrimitiveRecord* primitive = (PrimitiveRecord*) key;
		if (primitive->getValue() == attribute)
			return value;
	}
	// not found
	return NULL;
}

bool SdpRecordParser::hasServiceUuid16(unsigned int uuid) const {
	return hasService(UuidRecord(SDP_SIZE_2, uuid));
}

bool SdpRecordParser::hasServiceUuid32(unsigned int uuid) const {
	return hasService(UuidRecord(SDP_SIZE_4, uuid));
}

bool SdpRecordParser::hasServiceUuid128(const unsigned char* uuid) const {
	Buffer buffer(uuid, 16);
	return hasService(UuidRecord(SDP_SIZE_16, buffer));
}

bool SdpRecordParser::hasService(const UuidRecord& uuid) const {
	Record* attribute = findAttribute(SDP_ATTRIB_CLASS_ID_LIST);
	if (!attribute)
		return false;
	if (attribute->getType() != SDP_TYPE_DES)
		return false;
	ArrayRecord* array = (ArrayRecord*) attribute;
	for (int i = 0; i < array->getCount(); i++) {
		Record* r = array->getRecord(i);
		if (r->getType() != SDP_TYPE_UUID)
			continue;
		UuidRecord* uuidRec = (UuidRecord*) r;
		if (uuidRec->matches(uuid))
			return true;
	}
	return false;
}

bool SdpRecordParser::getRfcommChannel(int* channel) {
	Record* attribute = findAttribute(SDP_ATTRIB_PROTOCOL_DESCRIPTOR_LIST);
	if (!attribute)
		return false;
	UuidRecord l2cap(SDP_SIZE_2, L2CAP_PROTOCOL_UUID16);
	UuidRecord rfcomm(SDP_SIZE_2, RFCOMM_PROTOCOL_UUID16);
	PrimitiveRecord value;
	ScriptBlock blocks[] = {
		{ CHECK, DES },
			{ CHECK, DES },
				{ CHECK, UUID, &l2cap },
			{ CHECK, END },
			{ CHECK, DES },
				{ CHECK, UUID, &rfcomm },
				{ READ, PRIMITIVE, &value },
			{ CHECK, END },
		{ CHECK, END },
		{ FINISH }
	};

	if (!checkScript(blocks, attribute))
		return false;
	*channel = (int) value.getValue();
	return true;
}

void SdpRecordParser::visitRecords(AttributeVisitor* visitor, const Record* record) {
	if (!record)
		return;
	switch (record->getType()) {
	case SDP_TYPE_DES:
	case SDP_TYPE_DEA: {
		ArrayRecord* array = (ArrayRecord*) record;
		visitor->arrayStart(array);
		for (int i = 0; i < array->getCount(); i++)
			visitAttribute(visitor, array->getRecord(i));
		visitor->arrayEnd(array);
		}
		break;
	default:
		visitor->visitRecord(record);
		break;
	}
}

void SdpRecordParser::visitAttribute(AttributeVisitor* visitor, const Record* attribute) {
	visitRecords(visitor, attribute);
	visitor->done();
}


void SdpRecordParser::SerializeVisitor::arrayStart(const ArrayRecord* r) {
	events.push_back(r->getType() == SDP_TYPE_DES ? SdpRecordParser::DES : SdpRecordParser::DEA);
	records.push_back(r);
}
void SdpRecordParser::SerializeVisitor::arrayEnd(const ArrayRecord* r) {
	events.push_back(SdpRecordParser::END);
	records.push_back(r);
}
void SdpRecordParser::SerializeVisitor::visitRecord(const Record* r) {
	records.push_back(r);
	switch (r->getType()) {
	case SDP_TYPE_UINT:
	case SDP_TYPE_SINT:
	case SDP_TYPE_BOOL:
		events.push_back(PRIMITIVE);
		break;
	case SDP_TYPE_DES:
	case SDP_TYPE_DEA:
		break;
	case SDP_TYPE_NIL:
		events.push_back(NIL);
		break;
	case SDP_TYPE_STRING:
	case SDP_TYPE_URL:
	events.push_back(STRING);
		break;
	case SDP_TYPE_UUID:
		events.push_back(UUID);
		break;
	}
}

bool SdpRecordParser::checkScript(const ScriptBlock* block, const Record* record) {
	SerializeVisitor* serializer = new SerializeVisitor();
	visitAttribute(serializer, record);
	int elements = serializer->events.size();
	int index = 0;
	while (block->verb != FINISH && index < elements) {
		RecordType type = serializer->events[index];
		const Record* r = serializer->records[index];
		switch (block->verb) {
		case CHECK:
			if (type != block->type)
				return false;
			switch (type) {
			case UUID: {
				const UuidRecord* r1 = (const UuidRecord*) r;
				UuidRecord* r2 = (UuidRecord*) block->record;
				if (!r1->matches(*r2))
					return false;
				}
				break;
			case PRIMITIVE: {
				const PrimitiveRecord* p1 = (const PrimitiveRecord*) r;
				PrimitiveRecord* p2 = (PrimitiveRecord*) block->record;
				if (p1->getValue() != p2->getValue())
					return false;
				}
				break;
			case NIL:
			case END:
			case DES:
			case DEA:
				break;
			case STRING: {
				const StringRecord* s1 = (const StringRecord*) r;
				StringRecord* s2 = (StringRecord*) block->record;
				if (strcmp(s1->getString(), s2->getString()))
					return false;
				}
				break;
			}
			break;
		case READ:
			if (type != block->type)
				return false;
			switch (type) {
			case UUID: {
				const UuidRecord* r1 = (const UuidRecord*) r;
				UuidRecord* r2 = (UuidRecord*) block->record;
				*r2 = *r1;
				}
				break;
			case PRIMITIVE: {
				const PrimitiveRecord* p1 = (const PrimitiveRecord*) r;
				PrimitiveRecord* p2 = (PrimitiveRecord*) block->record;
				*p2 = *p1;
				}
				break;
			case NIL:
			case END:
			case DES:
			case DEA:
				break;
			case STRING: {
				const StringRecord* s1 = (const StringRecord*) r;
				StringRecord* s2 = (StringRecord*) block->record;
				s2->set(s1->getType(), s2->getString());
				}
				break;
			}
			break;
		default:
			break;
		}
		index++;
		block++;
	}
	delete serializer;
	if (block->verb != FINISH)
		return false;
	return true;
}




