/*
    Bantumi
    Copyright 2011 Martin Storsjo

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

    Martin Storsjo
    martin@martin.st
*/

package st.martin.bantumi;

import java.io.IOException;

import android.app.Activity;
import android.app.AlertDialog;
import android.app.Dialog;
import android.app.ProgressDialog;
import android.bluetooth.BluetoothDevice;
import android.content.Context;
import android.content.DialogInterface;
import android.content.Intent;
import android.content.SharedPreferences;
import android.opengl.GLSurfaceView;
import android.os.Bundle;
import android.text.InputType;
import android.view.KeyEvent;
import android.view.MotionEvent;
import android.view.Window;
import android.widget.EditText;
import android.widget.LinearLayout;

public class BantumiActivity extends Activity implements BluetoothSelectListener, ConnectionManager {
	private static final int mDefaultPort = 8471;
	
	private static boolean isLoaded = false;

	public static void load() {
		if (isLoaded)
			return;
		System.loadLibrary("bantumigl");
		isLoaded = true;
	}

	@Override
	protected void onCreate(Bundle savedInstanceState) {
		super.onCreate(savedInstanceState);
		requestWindowFeature(Window.FEATURE_NO_TITLE);

		load();
		init();
		SharedPreferences prefs = getPreferences(Context.MODE_PRIVATE);
		int level = prefs.getInt("level", -1);
		int num = prefs.getInt("num", -1);
		setLevelNum(level, num);
		
		try {
			mBluetoothManager = new BluetoothManager();
		} catch (Throwable t) {
		}

		mGLSurfaceView = new GLSurfaceView(this);
		mRenderer = new BantumiRenderer(this);

		mGLSurfaceView.setRenderer(mRenderer);
		setContentView(mGLSurfaceView);
	}

	@Override
	protected void onResume() {
		super.onResume();
		mGLSurfaceView.onResume();
	}

	@Override
	protected void onPause() {
		super.onPause();
		mGLSurfaceView.onPause();
	}

	@Override
	protected void onStop() {
		super.onStop();
		SharedPreferences.Editor prefs = getPreferences(Context.MODE_PRIVATE).edit();
		int level = getLevel();
		int num = getNum();
		prefs.putInt("level", level);
		prefs.putInt("num", num);
		prefs.commit();
		if (mSelectDialog != null)
			mSelectDialog.onStop();
	}

	@Override
	public boolean onTouchEvent(MotionEvent event) {
		if (mRenderer != null) {
			if (event.getAction() == MotionEvent.ACTION_DOWN) {
				handleClick((int) event.getX(), (int) event.getY());
				return true;
			}
		}
		return super.onTouchEvent(event);
	}

	@Override
	public boolean onKeyDown(int keyCode, KeyEvent event) {
		switch (keyCode) {
		case KeyEvent.KEYCODE_BACK:
			handleButton(KEY_ESC);
			return true;
		case KeyEvent.KEYCODE_DPAD_UP:
			handleButton(KEY_UP);
			return true;
		case KeyEvent.KEYCODE_DPAD_DOWN:
			handleButton(KEY_DOWN);
			return true;
		case KeyEvent.KEYCODE_DPAD_LEFT:
			handleButton(KEY_LEFT);
			return true;
		case KeyEvent.KEYCODE_DPAD_RIGHT:
			handleButton(KEY_RIGHT);
			return true;
		case KeyEvent.KEYCODE_DPAD_CENTER:
			handleButton(KEY_SELECT);
			return true;
		}
		return super.onKeyDown(keyCode, event);
	}

	public static final int MULTIPLAYER_DIALOG = 1;
	public static final int TCP_HOST_DIALOG = 2;
	public static final int TCP_JOIN_DIALOG = 3;
	public static final int WARNING_DIALOG = 4;
	public static final int TCP_DIALOG = 5;
	public static final int BLUETOOTH_DIALOG = 6;
	public static final int BLUETOOTH_SELECT_DIALOG = 7;
	
	public void showMultiplayerDialog() {
		runOnUiThread(new Runnable() {
			public void run() {
				if (mBluetoothManager != null)
					showDialog(MULTIPLAYER_DIALOG);
				else
					showDialog(TCP_DIALOG);
			}
		});
	}
	
	private String warningStr = "";
	public void showWarning(String str) {
		finishProgress();
		warningStr = str;
		runOnUiThread(new Runnable() {
			public void run() {
				showDialog(WARNING_DIALOG);
			}
		});
	}
	
	private ProgressDialog progressDialog;
	public void finishProgress() {
		if (progressDialog != null) {
			progressDialog.dismiss();
			progressDialog = null;
		}
	}
	public void startProgress(int titleId, int messageId) {
		String title = getResources().getString(titleId);
		String message = getResources().getString(messageId);
		progressDialog = ProgressDialog.show(this, title, message);
		progressDialog.setCancelable(true);
		progressDialog.setOnCancelListener(new DialogInterface.OnCancelListener() {
			public void onCancel(DialogInterface dialog) {
				handleButton(KEY_ESC);
			}
		});
	}
	
	@Override
	protected Dialog onCreateDialog(final int id) {
		Dialog dialog = null;
		AlertDialog.Builder builder = null;
		LinearLayout layout = null;
		SharedPreferences prefs = getPreferences(Context.MODE_PRIVATE);
		switch(id) {
		case MULTIPLAYER_DIALOG:
			builder = new AlertDialog.Builder(this);
			builder.setTitle(R.string.multiplayer_title)
			.setMessage(R.string.multiplayer_msg)
			.setCancelable(true)
			.setPositiveButton(R.string.tcpip_button, new DialogInterface.OnClickListener() {
				public void onClick(DialogInterface dialog, int which) {
					showDialog(TCP_DIALOG);
				}
			})
			.setNeutralButton(R.string.bluetooth_button, new DialogInterface.OnClickListener() {
				public void onClick(DialogInterface dialog, int which) {
					showDialog(BLUETOOTH_DIALOG);
				}
			})
			.setNegativeButton(R.string.cancel_button, new DialogInterface.OnClickListener() {
				public void onClick(DialogInterface dialog, int which) {
					handleButton(KEY_ESC);
				}
			})
			.setOnCancelListener(new DialogInterface.OnCancelListener() {
				public void onCancel(DialogInterface dialog) {
					handleButton(KEY_ESC);
				}
			});
			dialog = builder.create();
			break;
		case TCP_DIALOG:
		case BLUETOOTH_DIALOG:
			builder = new AlertDialog.Builder(this);
			builder.setTitle(id == TCP_DIALOG ? R.string.multiplayer_tcpip_title : R.string.multiplayer_bluetooth_title)
			.setMessage(R.string.hostjoin_msg)
			.setCancelable(true)
			.setPositiveButton(R.string.host_button, new DialogInterface.OnClickListener() {
				public void onClick(DialogInterface dialog, int which) {
					if (id == TCP_DIALOG)
						showDialog(TCP_HOST_DIALOG);
					else
						startBluetoothHost();
				}
			})
			.setNeutralButton(R.string.join_button, new DialogInterface.OnClickListener() {
				public void onClick(DialogInterface dialog, int which) {
					if (id == TCP_DIALOG)
						showDialog(TCP_JOIN_DIALOG);
					else
						startBluetoothJoin();
				}
			})
			.setNegativeButton(R.string.cancel_button, new DialogInterface.OnClickListener() {
				public void onClick(DialogInterface dialog, int which) {
					handleButton(KEY_ESC);
				}
			})
			.setOnCancelListener(new DialogInterface.OnCancelListener() {
				public void onCancel(DialogInterface dialog) {
					handleButton(KEY_ESC);
				}
			});
			dialog = builder.create();
			break;
		case TCP_JOIN_DIALOG:
			layout = new LinearLayout(this);
			layout.setOrientation(LinearLayout.VERTICAL);
			layout.setPadding(10, 10, 10, 10);
			final EditText host = new EditText(this);
			host.setHint(R.string.hostname_hint);
			host.setInputType(InputType.TYPE_CLASS_TEXT | InputType.TYPE_TEXT_FLAG_NO_SUGGESTIONS);
			final EditText port = new EditText(this);
			port.setHint(R.string.port_hint);
			port.setInputType(InputType.TYPE_CLASS_NUMBER);
			host.setText(prefs.getString("join_host", ""));
			port.setText(String.valueOf(prefs.getInt("join_port", mDefaultPort)));
			layout.addView(host);
			layout.addView(port);
			builder = new AlertDialog.Builder(this);
			builder.setTitle(R.string.multiplayer_tcpip_join_title)
			.setMessage(R.string.multiplayer_tcpip_join_msg)
			.setView(layout)
			.setCancelable(true)
			.setPositiveButton(R.string.connect_button, new DialogInterface.OnClickListener() {
				public void onClick(DialogInterface dialog, int which) {
					SharedPreferences.Editor prefs = getPreferences(Context.MODE_PRIVATE).edit();
					String hostname = host.getText().toString();
					int portValue = Integer.valueOf(port.getText().toString());
					prefs.putString("join_host", hostname);
					prefs.putInt("join_port", portValue);
					prefs.commit();
					doConnect(hostname, portValue);
					startProgress(R.string.connecting, R.string.connecting_msg);
				}
			})
			.setNegativeButton(R.string.cancel_button, new DialogInterface.OnClickListener() {
				public void onClick(DialogInterface dialog, int which) {
					handleButton(KEY_ESC);
				}
			})
			.setOnCancelListener(new DialogInterface.OnCancelListener() {
				public void onCancel(DialogInterface dialog) {
					handleButton(KEY_ESC);
				}
			});
			dialog = builder.create();
			break;
		case TCP_HOST_DIALOG:
			layout = new LinearLayout(this);
			layout.setOrientation(LinearLayout.VERTICAL);
			layout.setPadding(10, 10, 10, 10);
			final EditText hostPort = new EditText(this);
			hostPort.setHint(R.string.port_hint);
			hostPort.setInputType(InputType.TYPE_CLASS_NUMBER);
			hostPort.setText(String.valueOf(prefs.getInt("host_port", mDefaultPort)));
			layout.addView(hostPort);
			builder = new AlertDialog.Builder(this);
			builder.setTitle(R.string.multiplayer_tcpip_host_title)
			.setMessage(R.string.multiplayer_tcpip_host_msg)
			.setView(layout)
			.setCancelable(true)
			.setPositiveButton(R.string.host_button, new DialogInterface.OnClickListener() {
				public void onClick(DialogInterface dialog, int which) {
					SharedPreferences.Editor prefs = getPreferences(Context.MODE_PRIVATE).edit();
					int portValue = Integer.valueOf(hostPort.getText().toString());
					prefs.putInt("host_port", portValue);
					prefs.commit();
					doConnect(null, portValue);
					startProgress(R.string.waiting, R.string.waiting_msg);
				}
			})
			.setNegativeButton(R.string.cancel_button, new DialogInterface.OnClickListener() {
				public void onClick(DialogInterface dialog, int which) {
					handleButton(KEY_ESC);
				}
			})
			.setOnCancelListener(new DialogInterface.OnCancelListener() {
				public void onCancel(DialogInterface dialog) {
					handleButton(KEY_ESC);
				}
			});
			dialog = builder.create();
			break;
		case WARNING_DIALOG:
			builder = new AlertDialog.Builder(this);
			builder.setTitle(R.string.warning_title)
			.setMessage("")
			.setNegativeButton("Ok", new DialogInterface.OnClickListener() {
				public void onClick(DialogInterface dialog, int which) {
				}
			});
			dialog = builder.create();
			break;
		case BLUETOOTH_SELECT_DIALOG:
			if (mSelectDialog == null)
				mSelectDialog = new BluetoothSelectDialog(this, this);
			dialog = mSelectDialog.getDialog();
			break;
		default:
			break;
		}
		return dialog;
	}
	@Override
	protected void onPrepareDialog(int id, Dialog dialog) {
		if (id == WARNING_DIALOG) {
			AlertDialog ad = (AlertDialog) dialog;
			ad.setMessage(warningStr);
		}
		super.onPrepareDialog(id, dialog);
	}
	
	@Override
	protected void onActivityResult(int requestCode, int resultCode, Intent data) {
		if (requestCode != 0 && mBluetoothManager != null) {
			mBluetoothManager.onActivityResult(requestCode, resultCode, data);
		}
	}
	private BluetoothManager mBluetoothManager;
	private void startBluetoothHost() {
		mBluetoothManager.turnOnBluetooth(new Runnable() {
			public void run() {
				doBluetoothHost();
			}
		},
		new Runnable() {
			public void run() {
				handleButton(KEY_ESC);
			}
		}, this);
	}
	private void doBluetoothHost() {
		if (mConnectionThread != null)
			mConnectionThread.close();
		mConnectionThread = null;
		try {
			mConnectionThread = mBluetoothManager.doBluetoothHost(new Runnable() {
				public void run() {
					finishProgress();
					handleButton(KEY_ESC);
				}
			}, this, this);
		} catch (IOException ioe) {
			showWarning(ioe.toString());
			handleButton(KEY_ESC);
			return;
		}
		startProgress(R.string.waiting, R.string.waiting_msg);
	}
	public void bluetoothDeviceSelected(BluetoothDevice dev) {
		if (mConnectionThread != null)
			mConnectionThread.close();
		mConnectionThread = null;
		try {
			mConnectionThread = mBluetoothManager.connectTo(dev, this, this);
			startProgress(R.string.connecting, R.string.connecting_msg);
		} catch (IOException ioe) {
			showWarning(ioe.toString());
			handleButton(KEY_ESC);
		}
	}
	public void bluetoothSelectionCancelled() {
		handleButton(KEY_ESC);
	}
	private void startBluetoothJoin() {
		mBluetoothManager.turnOnBluetooth(new Runnable() {
			public void run() {
				showDialog(BLUETOOTH_SELECT_DIALOG);
			}
		}, new Runnable() {
			public void run() {
				handleButton(KEY_ESC);
			}
		}, this);
	}
	
	private ConnectionThread mConnectionThread;
	public void closeConnection() {
		if (mConnectionThread != null) {
			mConnectionThread.close();
			mConnectionThread = null;
		}
	}
	public void writeOutputData(byte[] buffer) {
		if (mConnectionThread != null)
			mConnectionThread.writeOutputData(buffer);
	}
	public native void handleInputData(byte[] buffer, int n);
	public native void initJavaConnection(boolean client);
	public native void setConnectionStatus(boolean working, String message);
	
	private GLSurfaceView mGLSurfaceView;
	private BantumiRenderer mRenderer;
	private BluetoothSelectDialog mSelectDialog;
	
	@SuppressWarnings("unused")
	private long mPrivData;

	private native void init();
	private native void handleClick(int x, int y);
	private native void setLevelNum(int level, int num);
	private native int getLevel();
	private native int getNum();
	private native void handleButton(int button);
	private native void doConnect(String host, int port);
	public native boolean shouldExit();
	private native void close();
	public void finalize() {
		close();
	}

	private static final int KEY_UP = 0;
	private static final int KEY_DOWN = 1;
	private static final int KEY_LEFT = 2;
	private static final int KEY_RIGHT = 3;
	private static final int KEY_SELECT = 4;
	@SuppressWarnings("unused")
	private static final int KEY_START = 5;
	private static final int KEY_ESC = 6;
}
